/* KallistiOS ##version##

   examples/dreamcast/pvr/yuv_converter/yuv.c
   Copyright (C) 2023 Andy Barajas

   This example demonstrates the use of TA's YUV converter.
*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <errno.h>

#include <dc/pvr.h>
#include <dc/maple.h>
#include <dc/maple/controller.h>

#include <malloc.h>
#include <arch/arch.h>
#include <arch/timer.h>

#define PVR_TEXTURE_WIDTH 512
#define PVR_TEXTURE_HEIGHT 512

static pvr_poly_hdr_t hdr;
static pvr_vertex_t vert[4];
static pvr_ptr_t converted_txr;

static unsigned char *y_plane;
static unsigned char *u_plane;
static unsigned char *v_plane;

extern uint8 romdisk[];
KOS_INIT_ROMDISK(romdisk);

static int load_image(void) {
    FILE *file = fopen("/rd/420.yuv", "rb");
    size_t read_size = 0;

    if (file == NULL) {
        printf("Could not open the file.\n");
        goto error;
    }

    y_plane = memalign(32, PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT);
    u_plane = memalign(32, PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT / 4);
    v_plane = memalign(32, PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT / 4);

    if (!y_plane || !u_plane || !v_plane) {
        perror("Could not allocate memory for yuv plane");
        goto error;
    }

    read_size = fread(y_plane, 1, PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT, file);
    if (read_size != PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT) {
        perror("Could not read y_plane completely");
        goto error;
    }

    read_size = fread(u_plane, 1, PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT / 4, file);
    if (read_size != PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT / 4) {
        perror("Could not read u_plane completely");
        goto error;
    }

    read_size = fread(v_plane, 1, PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT / 4, file);
    if (read_size != PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT / 4) {
        perror("Could not read v_plane completely");
        goto error;
    }

    fclose(file);
    return 0;

error:
    if (file)
        fclose(file);

    return 1;
}

static void setup_pvr(void) {
    pvr_poly_cxt_t cxt;

    if(!(converted_txr = pvr_mem_malloc(PVR_TEXTURE_WIDTH * PVR_TEXTURE_HEIGHT * 2))) {
        printf("Failed to allocate PVR memory!\n");
        //return NULL;
    }

    pvr_poly_cxt_txr(&cxt, PVR_LIST_OP_POLY, PVR_TXRFMT_YUV422 | PVR_TXRFMT_NONTWIDDLED, PVR_TEXTURE_WIDTH, PVR_TEXTURE_HEIGHT, converted_txr, PVR_FILTER_BILINEAR);
    pvr_poly_compile(&hdr, &cxt);

    //hdr.mode3 |= PVR_TXRFMT_STRIDE; // Was 0x02000000; which had one too many zeros. Should be 0x0200000

    vert[0].z     = vert[1].z     = vert[2].z     = vert[3].z     = 1.0f; 
    vert[0].argb  = vert[1].argb  = vert[2].argb  = vert[3].argb  = PVR_PACK_COLOR(1.0f, 1.0f, 1.0f, 1.0f);    
    vert[0].oargb = vert[1].oargb = vert[2].oargb = vert[3].oargb = 0;  
    vert[0].flags = vert[1].flags = vert[2].flags = PVR_CMD_VERTEX;         
    vert[3].flags = PVR_CMD_VERTEX_EOL;

    float ratio;
    int ul_x, ul_y, br_x, br_y;

    ratio = 640.0 / PVR_TEXTURE_WIDTH;
    ul_x = 0;
    br_x = (ratio * PVR_TEXTURE_WIDTH);
    ul_y = ((480 - ratio * PVR_TEXTURE_HEIGHT) / 2);
    br_y = ul_y + ratio * PVR_TEXTURE_HEIGHT;

    vert[0].x = ul_x;
    vert[0].y = ul_y;
    vert[0].u = 0.0;
    vert[0].v = 0.0;

    vert[1].x = br_x;
    vert[1].y = ul_y;
    vert[1].u = 1.0;
    vert[1].v = 0.0;

    vert[2].x = ul_x;
    vert[2].y = br_y;
    vert[2].u = 0.0;
    vert[2].v = 1.0;

    vert[3].x = br_x;
    vert[3].y = br_y;
    vert[3].u = 1.0;
    vert[3].v = 1.0;

    /* Set SQ to YUV converter. */
    PVR_SET(PVR_YUV_ADDR, (((unsigned int)converted_txr) & 0xffffff));
    // Divide texture width and texture height by 16 and subtract 1.
    // The actual values to set are 1, 3, 7, 15, 31, 63.
    PVR_SET(PVR_YUV_CFG_1, (((PVR_TEXTURE_HEIGHT / 16) - 1) << 8) | ((PVR_TEXTURE_WIDTH / 16) - 1));
    // Need to read once
    PVR_GET(PVR_YUV_CFG_1);
}

static void show_frame(void) {

    pvr_wait_ready();
    pvr_scene_begin();

    pvr_list_begin(PVR_LIST_OP_POLY);
    pvr_prim(&hdr, sizeof(pvr_poly_hdr_t));
    pvr_prim(&vert[0], sizeof(pvr_vertex_t));
    pvr_prim(&vert[1], sizeof(pvr_vertex_t));
    pvr_prim(&vert[2], sizeof(pvr_vertex_t));
    pvr_prim(&vert[3], sizeof(pvr_vertex_t));
    pvr_list_finish();

    pvr_scene_finish();
}

static void convert() {
    int i, j, index, x_blk, y_blk;

    unsigned char u_block[64] __attribute__((aligned(32)));
    unsigned char v_block[64] __attribute__((aligned(32)));
    unsigned char y_block[256] __attribute__((aligned(32)));

    for (y_blk = 0; y_blk < PVR_TEXTURE_HEIGHT; y_blk += 16) {
        for (x_blk = 0; x_blk < PVR_TEXTURE_WIDTH; x_blk += 16) {
            
            // Extract U
            for (j = 0; j < 8; ++j) {
                for (i = 0; i < 8; ++i) {
                    index = (y_blk / 2 + j) * (PVR_TEXTURE_WIDTH / 2) + (x_blk / 2 + i);
                    u_block[j * 8 + i] = u_plane[index];
                }
            }

            // Extract V
            for (j = 0; j < 8; ++j) {
                for (i = 0; i < 8; ++i) {
                    index = (y_blk / 2 + j) * (PVR_TEXTURE_WIDTH / 2) + (x_blk / 2 + i);
                    v_block[j * 8 + i] = v_plane[index];
                }
            }

            // Extract Y
            for (j = 0; j < 16; ++j) {
                for (i = 0; i < 16; ++i) {
                    index = (y_blk + j) * PVR_TEXTURE_WIDTH + (x_blk + i);
                    y_block[j * 16 + i] = y_plane[index];
                }
            }

            sq_cpy((void *)PVR_TA_YUV_CONV, (void *)u_block, 64);
            sq_cpy((void *)PVR_TA_YUV_CONV, (void *)v_block, 64);
            sq_cpy((void *)PVR_TA_YUV_CONV, (void *)y_block, 256);
        }

        // Send dummy mb
        //sq_set((void *)PVR_TA_YUV_CONV, 0, 384 * (32 - (PVR_TEXTURE_WIDTH >> 4)));
    }
}

static void __attribute__((__noreturn__)) wait_exit(void) {
    maple_device_t *dev;
    cont_state_t *state;

    printf("Press any button to exit.\n");

    for(;;) {
        dev = maple_enum_type(0, MAPLE_FUNC_CONTROLLER);

        if(dev) {
            state = (cont_state_t *)maple_dev_status(dev);

            if(state)   {
                if(state->buttons)
                    arch_exit();
            }
        }
    }
}

int main(int argc, char *argv[]) {
    pvr_init_defaults();

    load_image();

    setup_pvr();

    uint64_t startTime = timer_ns_gettime64();
    convert();
    printf("Took: %lld\n", timer_ns_gettime64() - startTime);

    show_frame();

    wait_exit();

    free(y_plane);
    free(u_plane);
    free(v_plane);
    pvr_mem_free(converted_txr);

    return 0;
}
